{% extends "base_debian.j2" %}
{% block SETUP_BENTO_ENVARS %}
SHELL [ "/bin/bash", "-eo", "pipefail", "-c" ]
{{ super() }}
{% endblock %}
{% block SETUP_BENTO_ENTRYPOINT %}
RUN rm -rf /var/lib/{apt,cache,log}
{{ super() }}
{% endblock %}
